/*
*       Written by Gaute Gamnes, NTNU
*       spring of 2014. 140521
*
*		Code borrowed and modified from several sources...,
*       Serial information:
* 		http://en.wikibooks.org/wiki/Serial_Programming/termios
* 		http://blog.eduardofleury.com/archives/2007/11/16
*
* 		compile with: g++ main.cpp -o programExe -pthread
*
*		UART1 needs to be enabled for this to work
*		Run command: echo BB-UART1 > /sys/devices/bone_capemgr.*BACKSLASHslots
*		BACKSLASH must be replaced with an actual backslash /
*		Run the command on the BBB itself to enable UART1 on pin
*		P9:24 (Tx), P9:26 (Rx).
*
*/

#include <termios.h>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>

using namespace std;

extern "C" {
	#include <pthread.h>
}

#define TTY_SERIAL_PORT "/dev/ttyO1" // Path to serial port, UART 1, UART1 pins are P9: TX(24), RX(26)

int init_serial_communication(char* port) {
	int tty_fd = 0; // The returned file handle for the device
	tty_fd = open(port, O_RDWR | O_NOCTTY); // | O_NDELAY); Could use delay?
											// O_RDWR - opens port for read and write
											// O_NOCTTY - The port never becomes the controlling terminal of the process
											// O_NDELAY - Non-blocking IO
	if(tty_fd == -1) {
		printf("Failed to open port\n");
		return tty_fd;
	}

	struct termios serial_options, original_serial_options;
	if(tcgetattr(tty_fd, &original_serial_options) == -1) { 	// Get current terminal settings
		printf("Error getting tty attributes %s - %s(%d).\n", tty_fd, strerror(errno), errno);
		return -1;
	}
	serial_options = original_serial_options; // Copy settings to new and play with them

	cfsetspeed(&serial_options, 9600); // Set input and output speed
	// 8N1
	serial_options.c_cflag &= ~PARENB;
	serial_options.c_cflag &= ~CSTOPB;
	serial_options.c_cflag &= ~CSIZE;
	serial_options.c_cflag |= CS8;
	// no flow control
	//serial_options.c_cflag &= ~CRTSCTS;
	serial_options.c_cflag |= CREAD | CLOCAL;  // turn on READ & ignore ctrl lines
	serial_options.c_iflag = 0;
	serial_options.c_lflag = 0;
	serial_options.c_oflag = 0;

	//Might not be needed
	//serial_options.c_cc[VMIN]  = 0; //Wait for an array of a certain size to enter buffer before moving on?
	//serial_options.c_cc[VTIME] = 1; //Timeout after 0.1 seconds

	// Apply settings
	if(tcsetattr(tty_fd, TCSANOW, &serial_options) == -1) {
		printf("Error setting tty attributes %s - %s(%d).\n", tty_fd, strerror(errno), errno);
		return -1;
	}
	// clear all flags on descriptor, enable direct IO
	if(fcntl(tty_fd, F_SETFL, FNDELAY)) {
			printf("Error setting tty non-blocking %s - %s(%d).\n", tty_fd, strerror(errno), errno);
			exit(EXIT_FAILURE);
	}

	return tty_fd;
}

void* receiveFromSerial(void* argument) {
	int tty_fd = *(reinterpret_cast<int*>(argument));

	/* Receive/send information from serial port */
	char rec_byte;
	while(1) {
		int rdChar = read(tty_fd, &rec_byte, 1);
		if(rdChar > 0) {
			cout << "Rec_byte: " << rec_byte << endl;
		}

	}
	/* Done receiving/sending from serial port */

	return 0;
}

void* sendToSerial(void* argument) {
	int tty_fd = *(reinterpret_cast<int*>(argument));

	while(1) {
		char input[256];
		cout << (">");
		fgets(input, 256, stdin);
		int size = strlen(input);
		cout << "Input: " << input << " size: " << size << endl;

		write(tty_fd, input, size);
	}

	return 0;
}

int main(int argc, char** argv) {
	int tty_fd = init_serial_communication(TTY_SERIAL_PORT);
	if(tty_fd == -1) {
		printf("Error opening serial port!\n");
		exit(EXIT_FAILURE);
	}

	pthread_t receive_t, send_t;

	int createThread1 = pthread_create(&receive_t, NULL, receiveFromSerial, reinterpret_cast<void*>(&tty_fd));
	if(createThread1 != 0) cout << "Error creating receive thread" << endl;
	int createThread2 = pthread_create(&send_t, NULL, sendToSerial, reinterpret_cast<void*>(&tty_fd));
	if(createThread2 != 0) cout << "Error creating send thread" << endl;
	pthread_join(receive_t, NULL);
	pthread_join(send_t, NULL);

	close(tty_fd);

	return 0;
}





