/*
 * findGPSlocationGivenPixel
 *
 *  Created on: 31. mai. 2014
 *      Author: Gaute Gamnes
 */


// This applications finds the GPS location of an object in the frame, given
// that the GPS location of the camera and its orientation and heading are known
/*
 * This application assumes that a webcamera is connected, i.e. find a webcam at capture(1)
 * and also that OpenCV is installed and the libraries are correctly linked.
 * 
 * The camera orientation and GPS location/height needs to be hardcoded into the program,
 * could be done in run-time if desired.
 */

#include <stdio.h>
#include <iostream>
#include <string>
#include <time.h>
#include <sstream>

extern "C" {
	#include <pthread.h>
}

#include <opencv2/core/core.hpp>
#include <opencv2/features2d/features2d.hpp>
#include <opencv2/nonfree/nonfree.hpp>
#include <opencv2/calib3d/calib3d.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/objdetect/objdetect.hpp>
#include <opencv2/contrib/contrib.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/stitching/stitcher.hpp>
using namespace std;
using namespace cv;

#define PI 3.14159265358979323846264

struct thread_data_t {
	Mat frame;
	int test;
	int pixelX;
	int pixelY;
}GLOB_thread_data_t;

void onMouse(int event, int x, int y, int, void*) {
	if( event != CV_EVENT_LBUTTONDOWN ) return;
	Point pt = Point(x,y);
	for(int i=0; i<10; i++) {
		cout << "x: " << pt.x << "\t y: " << pt.y << endl;
		GLOB_thread_data_t.pixelX = pt.x;
		GLOB_thread_data_t.pixelY = pt.y;
	}
}

void * getWebcamFrames(void * argument) {
	//thread_data_t *data = (reinterpret_cast<thread_data_t*>(argument));
	VideoCapture capture(1);
	capture.set(CV_CAP_PROP_FRAME_HEIGHT, 480);
	capture.set(CV_CAP_PROP_FRAME_WIDTH, 640);

	Mat frame;
	namedWindow("frame", CV_WINDOW_AUTOSIZE);

	// Rotation matrices etc, pi is declared as PI
	
	// -------------INPUT, camera orientation ---------------------------
	double phi = 1.3474889519; // om x
	double theta = 0; // om y
	double psi = PI/2 + 1.0847126821; // om z
	// ------------------------------------------------------------------
	
	Mat rotX = (Mat_<double>(3,3) << 	1,0,0,
										0, cos(phi), sin(phi),
										0, -sin(phi), cos(phi) );
	Mat rotY = (Mat_<double>(3,3) << 	cos(theta), 0, -sin(theta),
										0, 1, 0,
										sin(theta), 0, cos(theta) );
	Mat rotZ = (Mat_<double>(3,3) << 	cos(psi), sin(psi), 0,
										-sin(psi), cos(psi), 0,
										0, 0, 1 );
	Mat rotW2C = rotX * rotY * rotZ;
	//cout << "RotXYZ" << " " << rotXYZ << endl;
	Mat K = (Mat_<double>(3,3) << 	8.1331102645202134e+02, 0, 3.1950000000000000e+02,
									0, 8.1331102645202134e+02, 2.3950000000000000e+02,
									0, 0, 1 );

									
	// -------------INPUT, camera height and location lat long---------------------------
	Mat Cw = (Mat_<double>(3,1) <<	0,0,-16 );
	double cameraLatitudeCoordinate = 63.41856;
	double cameraLongitudeCoordinate = 10.40109;
	// -----------------------------------------------------------------------------------

	Mat IandCw = (Mat_<double>(3,4) <<	1,0,0, -Cw.at<double>(0,0),
										0,1,0, -Cw.at<double>(1,0),
										0,0,1, -Cw.at<double>(2,0) );
	Mat P = K * rotW2C * IandCw;
	//cout << "MAT:" << " " << P << endl;
	Mat Pwith3columnRemoved = (Mat_<double>(3,3) <<
			P.at<double>(0,0), P.at<double>(0,1), P.at<double>(0,3),
			P.at<double>(1,0), P.at<double>(1,1), P.at<double>(1,3),
			P.at<double>(2,0), P.at<double>(2,1), P.at<double>(2,3)  );

	// Calculate Xw homogeneous using the pixel Xi
	Mat Xi = (Mat_<double>(3,1) << 320, 240, 1 );
	Mat Xw = Pwith3columnRemoved.inv() * Xi;
	Xw = Xw / Xw.at<double>(2,0);
	cout << "MAT:" << " " << Xw << endl;

	// Declare to make sure they contain value
	GLOB_thread_data_t.pixelX = 0;
	GLOB_thread_data_t.pixelY = 0;

	int key = 0;
	while(key != 27) {
		capture >> frame;
		cv::circle(frame, Point(320,240), 5, CV_RGB(0,0,250), 2, 8);
		setMouseCallback("frame", onMouse, 0);
		//mat, point, radius, color, thickness, connectivitiy
		imshow("frame", frame);
		imwrite("locationResult.jpg", frame);

		Xi.at<double>(0,0) = GLOB_thread_data_t.pixelX;
		Xi.at<double>(1,0) = GLOB_thread_data_t.pixelY;
		Xw = Pwith3columnRemoved.inv() * Xi;
		Xw = Xw / Xw.at<double>(2,0);
		// Radius earth is assumed to be 6371, new LAT and LONG are calculated based on formulas found
		// http://stackoverflow.com/questions/7477003/calculating-new-longtitude-latitude-from-old-n-meters
		double newLatCoordinate = cameraLatitudeCoordinate + ( (Xw.at<double>(0,0)/1000) / 6371 ) * (180/PI);
		double newLongCoordinate = cameraLongitudeCoordinate +
				( (Xw.at<double>(1,0)/1000) / (6371*cos(PI*newLatCoordinate/180)) ) * (180/PI);
		cout << "World Coordinate: (" << Xw.at<double>(0,0) << ", " << Xw.at<double>(1,0) << "), New GPS lat/long: ("
				<< newLatCoordinate << ", " << newLongCoordinate << ")" << endl;

		key = waitKey(1);
	}
	pthread_exit(NULL);
	return 0;
}

int main(int argc, char** argv) {
	pthread_t getImageFromWebcam;

	std::cout.precision(8);

	int createThread1 = pthread_create(&getImageFromWebcam, NULL, getWebcamFrames, NULL);//reinterpret_cast<void*>(&thread_data_t));
	if(createThread1 != 0) exit(EXIT_FAILURE);

	pthread_join(getImageFromWebcam, NULL);

	return 0;
}
